from cmath import e
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers, initializers
from pysc2.agents import base_agent
from pysc2.env import sc2_env
from pysc2.lib import actions, features, units
import numpy as np

class NeuralNetwork():
    # Declare Constants used us PySC2
    _PLAYER_RELATIVE = features.SCREEN_FEATURES.player_relative.index
    _PLAYER_FRIENDLY = 1
    _PLAYER_NEUTRAL = 3
    _PLAYER_HOSTILE = 4

    #Move_screen actually moves the unit you have selected
    _MOVE_SCREEN = actions.FUNCTIONS.Move_screen.id
    _SELECT_ARMY = actions.FUNCTIONS.select_army.id
    _SELECT_POINT = actions.FUNCTIONS.select_point.id
    _ATTACK_SCREEN = actions.FUNCTIONS.Attack_screen.id
    _NO_OP = actions.FUNCTIONS.no_op.id

    #Unit Type constants
    _SCV = 45
    _MINERAL_FIELD = 341
    _VESPENE_GEYSER = 342
    _COMMAND_CENTER = 18
    _REFINERY = 20
    _SUPPLY_DEPOT = 19
    _MARINE = 48
    _CYBERNETICS_CORE = 72
    _NEXUS = 59
    _HATCHERY = 86

    def __init__ (self, layer_structure, output_actions):
        self.layer_structure = layer_structure
        self.output_actions = output_actions
        

    def initialize_model(self, input_size): #Currently this is run upon initialization of the object
        self.modelList = []
        #Add the input layer
        #Get the size of the input
        #inputVector = self.generateCurrentState(obs)
        self.modelList.append(keras.Input(shape=(input_size), ))
        #Add batch normalization of inputs
        self.modelList.append(keras.layers.BatchNormalization())
        #Add as many hidden layers as the user defines
        for layer_size in self.layer_structure: #self.layer_structure should be a list of integers that contains the number of neruons. We can make this more sophisticated when we want different activiation fucntions, etc
            self.modelList.append(layers.Dense(layer_size, kernel_initializer = initializers.TruncatedNormal()))
        #Add the output layer
        self.modelList.append(layers.Dense(4, activation = "softmax"))
        self.model = keras.Sequential(self.modelList)
    
    #def train_step(self): #This funciton will receive the state of the game and a reward value. It will then train the model for a step. 

    def generateCurrentState (self, obs): #This function is intended to generate an input state vector that can be utilized for training an interpretation of the network

        self.nnInputs = [] #Storage for inputs to the NN
        self.nnInputs.clear()

        mines = []
        vespian = []
        commandCenters = []
        marines = []
        enemies = []

        #If the build_SCV flag is set, a command center has already been selected and we need to tell it to build an SCV. None of the other code is required
        #if self.build_SCV == True:
        #    self.build_SCV = False
        #    return actions.FunctionCall(actions.FUNCTIONS.Train_SCV_quick.id, [0])

        for unit in obs.observation.raw_units:
            if unit.unit_type == self._MINERAL_FIELD:
                mines.append(unit)
            elif unit.unit_type == self._VESPENE_GEYSER or unit.unit_type == self._REFINERY:
                vespian.append(unit)
            elif unit.unit_type == self._COMMAND_CENTER:
                commandCenters.append(unit)
            elif unit.unit_type == self._MARINE and unit.alliance == self._PLAYER_FRIENDLY:
                marines.append(unit)
            elif unit.alliance == self._PLAYER_HOSTILE: #Commies
                enemies.append(unit)
        
        #Get enemy health
        numCommies = len(enemies)
        commieHealth = 0

        for commie in enemies:
            commieHealth += commie.health

        if numCommies > 0: #Do not implode the universe
            commieHealth /= numCommies

        #Determine who the enemy is
        enemyRaceTerrans = 0
        enemyRaceProtoss = 0
        enemyRaceZerg = 0
        enemyRaceUnknown = 1

        for unit in obs.observation.raw_units:
            if unit.alliance == self._PLAYER_HOSTILE:
                if unit.unit_type == self._COMMAND_CENTER:
                    enemyRaceTerrans = 1 
                    break
                elif unit.unit_type == self._NEXUS:
                    enemyRaceProtoss = 1
                    break
                elif unit.unit_type == self._HATCHERY:
                    enemyRaceZerg = 1
        if enemyRaceProtoss == 1 or enemyRaceTerrans == 1 or enemyRaceZerg == 1:
            enemyRaceUnknown = 0
        else:
            numCommies = 999
            commieHealth = 999

        #Get marine health
        numMarines = len(marines)
        marineHealth = 0

        for marine in marines:
            marineHealth += marine.health
        
        if numMarines > 0: #Do not implode the universe
            marineHealth /= numMarines #Make to normalize th value for the network

        
            
        #Define some inputs
        numMines = len(mines)
        numVespian = len(vespian)
        numCCs = len(commandCenters)
        numSCVs = obs.observation.player.food_used #Starcraft represents the total units in the supply depot as food_used for some reason
        #Find the number of idle SCVs
        numSupplyDepots = 0
        numIdleSCVs = 0
        timestep = obs.observation.game_loop[0]

        #Search through all units to determine characterisitcs of the current state
        for unit in obs.observation.raw_units:
            if unit.unit_type == self._SCV:
                if unit.order_length == 0:
                    numIdleSCVs += 1
            if unit.unit_type == self._SUPPLY_DEPOT:
                numSupplyDepots += 1

        percentageGarrisonFull = obs.observation.player.food_used / obs.observation.player.food_cap #Food used and food cap are the supply depot capacities that are mislabeled. 
        #Clear the previous state
        self.nnInputs.clear()

        #Build the input vector
        self.nnInputs.append(enemyRaceTerrans)
        self.nnInputs.append(enemyRaceProtoss)
        self.nnInputs.append(enemyRaceZerg)
        self.nnInputs.append(enemyRaceUnknown)
        self.nnInputs.append(numCommies)
        self.nnInputs.append(commieHealth)
        self.nnInputs.append(marineHealth) 
        self.nnInputs.append(numMarines) 
        self.nnInputs.append(timestep)

    ''' Train model
        This function will be called at each step of the game. It will receive a History of the Training Data
    '''

    def trainModel (self, history):
        x_train = history.modelInputs.astype("float32")
        y_train = history.modelOutputs.astype("float32")

        self.model.compile(
            optimizer = keras.optimizers.SGD(learning_rate = 0.01, momentum = 0.25),
            #loss = keras.losses.CategoricalCrossentropy(),
            loss = keras.losses.MeanAbsoluteError(),
            metrics=["accuracy"] 
            )
        
        outputData = self.model.fit(x_train, y_train, batch_size = 32, epochs = 100)
        self.model.summary()
        #prediction = self.model.predict(x_train[:3])
        #print("Hello World")

    def predictMove(self, obs): #gameState is a NumPy array with the current state of the game that the model was trained with
        _GAME_ACTIONS = ["home",
                              "enemy",
                              "retreat-horizontal",
                              "retreat-vertical"]

        self.generateCurrentState(obs) #Provide the observation, get the state in the form of a list

        #Convert to NumPy array
        x_dim = len(self.nnInputs) 
        stateArray = np.empty([1, x_dim])
        for idx in range(len(self.nnInputs)):
            stateArray[0][idx] = self.nnInputs[idx]

        prediction = self.model.predict(stateArray) 
        actionID = np.argmax(prediction[0])
        print("NN Selected Action: " + _GAME_ACTIONS[actionID])
        for gameIdx in range(0, len(_GAME_ACTIONS)):
            print("Game action " + _GAME_ACTIONS[gameIdx] + " probability: " + str(prediction[0][gameIdx]))
        return actionID

    def saveModel (self):
        self.model.save("my_model")







