'''
This class is intended to abstract the starting script for each mission, 
where several supply depots and barracks are created
'''
from ScreenSectors import ScreenSectors
from GameState import GameState
from BuildSupplyDepot import BuildSupplyDepot
from BuildMarineBarracks import BuildMarineBarracks
from ScenarioRecord import ScenarioRecord
from MacroAction import MacroAction 
from pysc2.lib import actions, features, units
from BuildMarine import BuildMarine
import random as rand
from MoveMarine import MoveMarine


class RunMission():
    _COMMAND_CENTER = 18
    _MARINE_POSSIBLE_MOVES = ["home",
                              "enemy",
                              "retreat-horizontal",
                              "retreat-vertical"]

    def __init__ (self, timesteps, numBarracks, numSupplyDepots, env, agent):
        self.agent = agent
        self.record = ScenarioRecord()
        self.gameState = GameState(timesteps[0])
        self.env = env
        self.macroAction = MacroAction()
        self.startMission(timesteps, numBarracks, numSupplyDepots, env)

    def getScenarioRecord (self):
        return self.record

    def clearScenarioRecord (self):
        self.record.clearRecord()

    def startMission (self, timesteps, numBarracks, numSupplyDepots, env):
        _NUM_BARRACKS = numBarracks
        _NUM_SUPPLY_DEPOTS = numSupplyDepots

        #actionSpaceCounter = 0 #Limit how many actions can be issued
        self.screenSectors = ScreenSectors(64)
        self.screenSectors.createSectors(4, 4)
        
        
        #Maintains a record for each epoch
        

        #Find the command center
        for unit in timesteps[0].observation.raw_units:
            if unit.unit_type == self._COMMAND_CENTER:
                command_center = unit
                break

        #Determine which sector contains the Command Center and output the min, max grid coordinates
        self.friendDims = self.screenSectors.getFriendlyArea(command_center)

        #Constants for the initial game sequence
        _NUM_BARRACKS = 2
        _NUM_SUPPLY_DEPOTS = 4
        _BASE_COMPLETE = False
        
        _AGENT_SLOW = 1 #Only take an action once every _AGENT_SLOW steps

        #Create a macro action object. This will be used to move all marines at once and nothing else
        armyMove = MacroAction()
        gameCounter = 0
        #Initialize the first set of training epochs

        #Should break out of this loop as soon as the appropriate number of barracks and supply depots have been built
        while True:
            #This experiment will only have the Neural Network learn how to fight. 
            #I will script building the Barracks and Supply Depots. The AI will start issuing actions as soon as the first marine is built.

            #Scripted Actions
            self.gameState.stateDict = self.gameState.updateState(timesteps[0]).copy()
            #Adding this check to stop the SCVs from building after attack
            if self.gameState.getNumBarracks() >= _NUM_BARRACKS and self.gameState.getNumSupplyDepots() >= _NUM_SUPPLY_DEPOTS:
                _BASE_COMPLETE = True
                break
            
            minerals = self.gameState.getMinerals()
            num_supply_depots_built = self.gameState.getNumSupplyDepots()
            num_barracks_built = self.gameState.getNumBarracks()
                    
            #Build _NUM_SUPPLY_DEPOTS First
            if num_supply_depots_built < _NUM_SUPPLY_DEPOTS and self.gameState.canBuildSupplyDepot() == True and _BASE_COMPLETE == False:
                actionObj = BuildSupplyDepot(timesteps[0], self.friendDims)
                num_supply_depots_built += 1
            elif num_barracks_built < _NUM_BARRACKS and self.gameState.canBuildBarracks() == True and _BASE_COMPLETE == False:
                actionObj = BuildMarineBarracks(timesteps[0], self.friendDims)
                num_barracks_built += 1     
            else:
                actionObj = "No Action"

            if actionObj != "No Action":
                step_actions = [self.agent.step(timesteps[0], actionObj, self.macroAction, self.screenSectors)] #Need to complete the first step of the action
            else:
                step_actions = [actions.RAW_FUNCTIONS.no_op()]
            timesteps = env.step(step_actions)

            if timesteps[0].last():
                self.record.normalizeReward(timesteps[0].reward) #Come up with the reward for all step actions
                self.record.assignReward()
                #history.addHistoryRecord(self.record)
                _BASE_COMPLETE = False
                break

    '''
    This function will run the mission, where the barracks is building marines and 
    we are randomly selecting a location where the marines should move.
        
    agent == PyAgent object. Needed to determine if we are training or not
    model == Needed to make predictions if we are training. Set to None if not training
    env == Need the game environment to make a timestep occur
    '''

    def runMission (self, timesteps, model):
        armyMove = MacroAction()
        moveOrMarine = "marine"
        marineCount = 0
        _ACTION_SPACING = 3
        actionSpaceCounter = _ACTION_SPACING

        _MARINE_POSSIBLE_MOVES = ["home",
                              "enemy",
                              "retreat-horizontal",
                              "retreat-vertical"]

        while True:
                    #This function assumes the base has already been built and we are only building and moving marines

                    self.gameState.stateDict = self.gameState.updateState(timesteps[0]).copy()
                    minerals = self.gameState.getMinerals()
                    
                    if self.gameState.canMoveMarine() and moveOrMarine == "move":
                        if armyMove.isNewOrder() == True:
                            model.generateCurrentState(timesteps[0])
                            if actionSpaceCounter == 0:
                                armyMove.generateAction(self.agent.getTrianing(), model, timesteps[0])
                                actionSpaceCounter = _ACTION_SPACING
                            else:
                                actionSpaceCounter -= 1
                            self.record.addStep(model.nnInputs, armyMove.getAction(), True)
                            armyMove.setMarineTags(timesteps[0])
                            #I want to move all marines before we do anything else
                                       
                        marineTag = armyMove.getMarineTag()
                        if marineTag == None:
                            actionObj = "No Action"
                        else:
                            actionObj = MoveMarine(timesteps[0], self.friendDims, self.gameState.getCCLocation(), _MARINE_POSSIBLE_MOVES[armyMove.getAction()], marineTag)
                        if len(armyMove.marineTags) == 0:
                            moveOrMarine = "marine" 
                       
                    elif self.gameState.canBuildMarine() and moveOrMarine == "marine":
                        if len(timesteps[0].observation.production_queue) >= 3 or self.gameState.getNumMarines() > 10:
                            if rand.random() < 0.4:
                                moveOrMarine = "move"
                        actionObj = BuildMarine(timesteps[0], self.friendDims)
                    else:
                        actionObj = "No Action"

                    if actionObj != "No Action":
                        step_actions = [self.agent.step(timesteps[0], actionObj, self.macroAction, self.screenSectors)] #Need to complete the first step of the action
                    else:
                        step_actions = [actions.RAW_FUNCTIONS.no_op()]
                    timesteps = self.env.step(step_actions)

                    if timesteps[0].last():
                        self.record.normalizeReward(timesteps[0].reward) #Come up with the reward for all step actions
                        self.record.assignReward()
                        #history.addHistoryRecord(record)
                        _BASE_COMPLETE = False
                        break